


import torch
import torch.nn as nn
import torch.nn.functional as F

torch.manual_seed(42)
import brevitas.nn as qnn
from brevitas.quant import Int32Bias

qw = 8
qa = 8


class Quantized_S2_Block(nn.Module):
    """ S2 ConvBlock used in Temporal Convolutions
        -> CONV -> BN -> RELU -> CONV -> BN -> (+) -> RELU
        |_______-> CONV -> BN -> RELU ->________|
    """

    def __init__(self, in_ch, out_ch):

        super(Quantized_S2_Block, self).__init__()
        # self.conv0_d = nn.Conv2d(in_channels=in_ch, out_channels=in_ch, kernel_size=(1, 9), stride=2, padding=(0, 4),
        #                          bias=False, groups=in_ch)
        self.conv0_d = qnn.QuantConv2d(in_channels=in_ch, out_channels=out_ch, kernel_size=(1, 9), stride=2,
                                       padding=(0, 4), bias=False, weight_bit_width=qw, bias_quant=Int32Bias)
        self.relu0 = qnn.QuantReLU(bit_width=qw, return_quant_tensor=True)


        self.conv1_d = qnn.QuantConv2d(in_channels=out_ch, out_channels=out_ch, kernel_size=(1, 9), stride=1,
                                       padding=(0, 4), bias=False, weight_bit_width=qw, bias_quant=Int32Bias)
        self.relu2 = qnn.QuantReLU(bit_width=qw, return_quant_tensor=True)




    def forward(self, x):
        # self.is_quant_act = is_quant_act
        out = self.forward_train(x)

        return out

    def forward_train(self, x):

        out = self.conv0_d(x)
        out = self.relu0(out)


        out = self.conv1_d(out)
        out = self.relu2(out)


        return out


class QuantizedTCResNet8(nn.Module):
    """ TC-ResNet8 implementation.

        Input dim:  [batch_size x n_mels x 1 x 101]
        Output dim: [batch_size x n_classes]

        Input -> Conv Layer -> (S2 Block) x 3 -> Average pooling -> FC Layer -> Output
    """

    def __init__(self, k, n_mels, n_classes):
        super(QuantizedTCResNet8, self).__init__()
        self.mode = "train"
        self.is_quant_act = True

        self.quant_inp = qnn.QuantIdentity(bit_width=qa, return_quant_tensor=True)


        self.conv_block_d = qnn.QuantConv2d(in_channels=n_mels, out_channels=int(16 * k), kernel_size=(1, 3), stride=1,
                                            bias=False,  weight_bit_width=qw, padding=(0, 1), bias_quant=Int32Bias
                                            )
        self.relu0 = qnn.QuantReLU(bit_width=qw,return_quant_tensor=True)

        # S2 Blocks
        self.s2_block0 = Quantized_S2_Block(int(16 * k), int(24 * k))
        self.s2_block1 = Quantized_S2_Block(int(24 * k), int(32 * k))
        self.s2_block2 = Quantized_S2_Block(int(32 * k), int(48 * k))



        self.fc = qnn.QuantLinear(int(48 * k), n_classes, bias=False, weight_bit_width=qw)
        # self.fc = qnn.QuantLinear(int(24 * k), n_classes, bias=False, weight_bit_width=qw)
        self.relu2 = qnn.QuantReLU(bit_width=qw, return_quant_tensor=True)
        # self.fc = qnn.QuantConv2d(in_channels=int(48 * k), out_channels=n_classes, kernel_size=1, padding=0,
        #                                     bias=True, weight_bit_width=qw, bias_quant=Int32Bias)


    def forward(self, x):
        # print("nn input shape: ",x.shape)

        # if (self.mode == "train"):
        out = self.forward_path_train(x)
        return out
        # elif (self.mode == "eval"):
            # out = self.forward_path_eval(x)
            # return out.view(out.shape[0], -1)

    def forward_path_train(self, x):
        x = self.quant_inp(x)
        out = self.conv_block_d(x)
        out = self.relu0(out)
        # out = self.conv_block_p(out)
        # out = self.relu1(out)

        out = self.s2_block0(out)
        out = self.s2_block1(out)
        out = self.s2_block2(out)
        # out = F.max_pool2d(out, kernel_size=(1, 51), stride=1)
        out = F.max_pool2d(out, kernel_size=(1, 13), stride=1)
        out = out.reshape(out.shape[0], -1)
        out = self.fc(out)
        out = self.relu2(out)
        # out = F.softmax(out, dim=1)
        return out

    def save(self, is_onnx=0, name="saved_model/TCResNet8"):
        if (is_onnx):
            dummy_input = torch.randn(16, 30, 1, 101)
            torch.onnx.export(self, dummy_input, "TCResNet8.onnx", verbose=True, input_names=["input0"],
                              output_names=["output0"])
        else:
            torch.save(self.state_dict(), name)

            # f_save = open('activation.pkl', 'wb')
            # pickle.dump(min_max_act, f_save)
            # f_save.close()

    def load(self, name="saved_model/TCResNet8"):
        self.load_state_dict(torch.load(name, map_location=lambda storage, loc: storage))


if __name__ == "__main__":



    model = QuantizedTCResNet8(1, 40, 10)
    model.load("weight.pt")
    model.eval()
    x = torch.randn(1, 40, 1, 101)
    out = model(x)
